from __future__ import annotations

from copy import deepcopy
from typing import Any, Callable, Dict, Generator, List, Optional, Sequence, Type, Union, get_type_hints

from pydantic import BaseModel, ConfigDict, Field
from pydantic.main import create_model
from pypika import Order, Query
from pypika.queries import QueryBuilder
from pypika.terms import Criterion, Parameter

from p3orm.core import dialect, driver, querybuilder
from p3orm.exceptions import (
    InvalidRelationship,
    MissingPrimaryKey,
    MissingRelationship,
    MissingTablename,
    MultiplePrimaryKeys,
    MultipleResultsReturned,
    NoResultsReturned,
)
from p3orm.fields import UNLOADED, RelationshipType, _PormField, _Relationship
from p3orm.types import Model
from p3orm.utils import is_optional, paramaterize, with_returning

FetchType = Sequence[Sequence[_Relationship]]


class TableInstance(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        populate_by_name=True,
        frozen=False,
    )

    __pydantic_complete__ = False

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, BaseModel):
            return (
                self.__dict__ == other.__dict__
                and self.__pydantic_private__ == other.__pydantic_private__
                and self.__pydantic_extra__ == other.__pydantic_extra__
            )
        else:
            return NotImplemented  # delegate to the other item in the comparison


class Table:
    __tablename__: str

    class Meta:
        meta_table: bool = False

    _model_factory: type[BaseModel]

    #
    # Magic
    #
    def __new__(cls: Union[Type[Model], Table], /, **_create_fields: Dict[str, Any]) -> Model:
        create_fields = deepcopy(_create_fields)

        for relationship_name in cls._relationship_map():
            create_fields[relationship_name] = UNLOADED()

        return cls._create_model_factory()(**create_fields)

    @classmethod
    def _create_model_factory(cls: Table) -> Type[BaseModel]:
        if hasattr(cls, "_model_factory") and cls._model_factory:
            return cls._model_factory

        class _TableInstance(TableInstance):
            ...

        factory_model_kwargs = {}
        for field_name, field in cls._field_map().items():
            field: _PormField

            default = ...
            if field.autogen or is_optional(field._type):
                default = None

            factory_model_kwargs[field_name] = (field._type, Field(default, alias=field.column_name))

        for relationship_name in cls._relationship_map():
            factory_model_kwargs[relationship_name] = (UNLOADED, None)

        factory = create_model(cls.__name__, __base__=_TableInstance, **factory_model_kwargs)
        cls._model_factory = factory
        return cls._model_factory

    def __init_subclass__(cls) -> None:
        # raise if concrete table without tablename
        if (not hasattr(cls, "__tablename__") or cls.__tablename__ is None) and cls.Meta.meta_table == False:
            raise MissingTablename(f"{cls.__name__} must define a __tablename__ property")

        # don't calculate for meta tables
        elif cls.Meta.meta_table and not hasattr(cls, "__tablename__"):
            return

        fields = cls._fields()

        num_pkeys = len([f for f in fields if f.pk])
        if num_pkeys == 0:
            raise MissingPrimaryKey(f"{cls.__name__} must have 1 primary key field")

        elif num_pkeys > 1:
            raise MultiplePrimaryKeys(f"{cls.__name__} has more than 1 primary key field (check parent classes)")

        cls._create_model_factory()

    @classmethod
    def __get_validators__(cls) -> Generator[Callable, None, None]:
        yield cls.__validate

    @classmethod
    def __validate(cls, v: Model | Any) -> Model:
        assert (
            hasattr(v.__class__, "schema") and v.__class__.schema() == cls._create_model_factory().schema()
        ), f"must be a valid {cls.__name__}"
        return v

    #
    # Introspective methods
    #
    @classmethod
    def _relationship_map(cls) -> Dict[str, _Relationship]:
        return {
            name: relationship for name in cls.__dict__ if isinstance(relationship := getattr(cls, name), _Relationship)
        }

    @classmethod
    def _relationship_field_name(cls, relationship: _Relationship) -> str:
        for field_name, _relationship in cls._relationship_map().items():
            if relationship == _relationship:
                return field_name

        raise MissingRelationship(f"_Relationship {relationship} does not exist on {cls.__name__}")

    @classmethod
    def _fields(cls, exclude_autogen: Optional[bool] = False) -> List[_PormField]:
        fields: List[_PormField] = []
        for table in cls.__mro__:
            for field_name in table.__dict__:
                if isinstance(field := getattr(table, field_name), _PormField):
                    if field.column_name is None:
                        field.column_name = field_name
                        field.name = field_name
                    field.field_name = field_name
                    fields.append(field)

        if exclude_autogen:
            fields = [f for f in fields if not f.autogen]

        return fields

    @classmethod
    def _field_map(cls, exclude_autogen: Optional[bool] = False) -> Dict[str, _PormField]:
        fields = cls._fields(exclude_autogen)
        return {f.field_name: f for f in fields}

    @classmethod
    def _db_values(cls, item: Model, exclude_autogen: Optional[bool] = False) -> List[Any]:
        return [getattr(item, field.field_name) for field in cls._fields(exclude_autogen=exclude_autogen)]

    @classmethod
    def _primary_key(cls) -> Optional[_PormField]:
        for field in cls._fields():
            if field.pk:
                return field
        return None

    @classmethod
    def _field(cls, field_name: str) -> _PormField:
        return cls._field_map()[field_name]

    @classmethod
    def is_type(cls, obj: Any) -> bool:
        return obj.__class__ == cls._model_factory

    #
    # Shortcuts
    #
    @classmethod
    def from_(cls) -> QueryBuilder:
        return querybuilder().from_(cls.__tablename__)

    @classmethod
    def select(cls, to_select: str = "*") -> QueryBuilder:
        return cls.from_().select(to_select)

    @classmethod
    def delete(cls) -> QueryBuilder:
        return querybuilder().delete().from_(cls.__tablename__)

    @classmethod
    def update(cls) -> QueryBuilder:
        return querybuilder().update(cls.__tablename__)

    #
    # Queries
    #
    @classmethod
    async def insert_one(
        cls: Union[Type[Model], Table],
        /,
        item: Model,
        *,
        prefetch: FetchType = None,
    ) -> Model:
        query: QueryBuilder = Query.into(cls.__tablename__).columns(cls._fields(exclude_autogen=True))

        query_args = cls._db_values(item, exclude_autogen=True)
        query_params = [Parameter(f"${i + 1}") for i in range(len(query_args))]

        query = query.insert(query_params)

        inserted: Optional[Model] = await driver().fetch_one(cls, with_returning(query), query_args)

        if prefetch and inserted:
            [inserted] = await cls.fetch_related([inserted], prefetch)

        return inserted

    @classmethod
    async def insert_many(
        cls: Union[Type[Model], Table],
        /,
        items: List[Model],
        *,
        prefetch: FetchType = None,
    ) -> List[Model]:
        if not items:
            return []

        columns = cls._fields(exclude_autogen=True)
        columns_count = len(columns)
        query: QueryBuilder = Query.into(cls.__tablename__).columns(*columns)
        query_args: List[Any] = []

        for page, item in enumerate(items):
            query_args += cls._db_values(item, exclude_autogen=True)
            query_params = [Parameter(f"${page * columns_count + offset + 1}") for offset in range(columns_count)]
            query = query.insert(query_params)

        inserted = await driver().fetch_many(cls, with_returning(query), query_args)

        if prefetch and inserted:
            inserted = await cls.fetch_related(inserted, prefetch)

        return inserted

    @classmethod
    async def fetch_first(
        cls: Union[Type[Model], Table],
        /,
        criterion: Criterion,
        *,
        prefetch: FetchType = None,
    ) -> Optional[Model]:
        paramaterized_criterion, query_args = paramaterize(criterion, dialect=dialect())
        query: QueryBuilder = cls.select().where(paramaterized_criterion)
        query = query.limit(1)

        result = await driver().fetch_one(cls, query.get_sql(), query_args=query_args)

        if result and prefetch:
            [result] = await cls.fetch_related([result], prefetch)

        return result

    @classmethod
    async def fetch_one(
        cls: Union[Type[Model], Table],
        /,
        criterion: Criterion,
        *,
        prefetch: FetchType = None,
    ) -> Model:
        paramaterized_criterion, query_args = paramaterize(criterion, dialect=dialect())

        query: QueryBuilder = cls.select().where(paramaterized_criterion)
        query = query.limit(2)

        results = await driver().fetch_many(cls, query.get_sql(), query_args)

        if len(results) > 1:
            raise MultipleResultsReturned(f"Multiple {cls.__name__} were returned when only one was expected")
        elif len(results) == 0:
            raise NoResultsReturned(f"No {cls.__name__} were returned when one was expected")

        result = results[0]

        if prefetch:
            [result] = await cls.fetch_related([result], prefetch)

        return result

    @classmethod
    async def fetch_all(
        cls: Union[Type[Model], Table],
        /,
        criterion: Criterion = None,
        *,
        order: Union[_PormField, List[_PormField]] = None,
        by: Order = Order.asc,
        limit: int = None,
        prefetch: FetchType = None,
    ) -> List[Model]:
        query: QueryBuilder = cls.select()

        query_args = None
        if criterion:
            parameterized_criterion, query_args = paramaterize(criterion, dialect=dialect())
            query = query.where(parameterized_criterion)

        if order:
            order_fields = order if isinstance(order, list) else [order]
            query = query.orderby(*[o.column_name for o in order_fields], order=by)

        if limit:
            query = query.limit(limit)

        results = await driver().fetch_many(cls, query.get_sql(), query_args)

        if prefetch:
            results = await cls.fetch_related(results, prefetch)

        return results

    @classmethod
    async def update_one(
        cls: Union[Type[Model], Table],
        /,
        item: Model,
        *,
        prefetch: FetchType = None,
    ) -> Model:
        query: QueryBuilder = querybuilder().update(cls.__tablename__)

        pk = cls._primary_key()
        parameterized_criterion, query_args = paramaterize(pk == getattr(item, pk.field_name), dialect=dialect())

        for i, field in enumerate(cls._fields()):
            field: _PormField
            query = query.set(field.column_name, Parameter(f"${i + 2}"))
            query_args.append(getattr(item, field.field_name))

        query = query.where(parameterized_criterion)

        updated = await driver().fetch_one(cls, with_returning(query), query_args)

        if prefetch and updated:
            [updated] = await cls.fetch_related([updated], prefetch)

        return updated

    @classmethod
    async def delete_where(cls: Union[Type[Model], Table], /, criterion: Criterion) -> List[Model]:
        query: QueryBuilder = querybuilder().delete()
        query = query.from_(cls.__tablename__)

        parameterized_criterion, query_args = paramaterize(criterion, dialect=dialect())

        query = query.where(parameterized_criterion)
        return await driver().fetch_many(cls, with_returning(query), query_args)

    @classmethod
    async def _load_relationships_for_items(
        cls: Union[Type[Model], Table],
        /,
        items: List[Union[Model, BaseModel]],
        _relationships: FetchType,
    ) -> List[Model]:
        """Loads relationships, updates items in place"""
        relationship_map = cls._relationship_map()
        type_hints = get_type_hints(cls)

        for relationships in _relationships:
            if not relationships:
                continue

            relationship = relationships[0]

            relationship_table: Type[Table] = None
            relationship_field_name: str = None
            for n, r in relationship_map.items():
                if r == relationship:
                    relationship_table = (
                        type_hints.get(n)
                        if relationship.relationship_type == RelationshipType.foreign_key
                        else type_hints.get(n).__args__[0]
                    )

                    relationship_field_name = n

            if not relationship_table:
                raise InvalidRelationship(f"Relationship {relationship} doesn't exist on {cls}")

            keys = (
                [getattr(item, relationship.self_column) for item in items]
                if relationship.relationship_type == RelationshipType.foreign_key
                else [getattr(item, relationship.self_column) for item in items]
            )

            paramaterized_criterion, query_args = paramaterize(
                relationship_table._field(relationship.foreign_column).isin(keys),
                dialect=dialect(),
            )
            sub_items_query: QueryBuilder = relationship_table.select().where(paramaterized_criterion)

            sub_items = await driver().fetch_many(relationship_table, sub_items_query.get_sql(), query_args)

            if relationship.relationship_type == RelationshipType.foreign_key:
                sub_items_map = {getattr(i, relationship.foreign_column): i for i in sub_items}

                for item in items:
                    foreign_key = getattr(item, relationship.self_column)

                    # Remove unloaded relationship as it has been fetched
                    if foreign_key is None:
                        setattr(item, relationship_field_name, None)
                        continue

                    setattr(item, relationship_field_name, sub_items_map[foreign_key])

            else:
                items_map = {getattr(i, relationship.self_column): i for i in items}

                # Attach all loaded sub_items to appropriate item
                for sub_item in sub_items:
                    foreign_key = getattr(sub_item, relationship.foreign_column)
                    item = items_map[foreign_key]

                    if isinstance(getattr(item, relationship_field_name), UNLOADED):
                        setattr(item, relationship_field_name, [sub_item])
                    else:
                        getattr(item, relationship_field_name).append(sub_item)

                # Remove UNLOADED relationships as they have now been loaded
                for item in items:
                    if isinstance(getattr(item, relationship_field_name), UNLOADED):
                        setattr(item, relationship_field_name, [])

            await relationship_table._load_relationships_for_items(sub_items, [relationships[1:]])

    @classmethod
    async def fetch_related(
        cls: Union[Type[Model], Table],
        /,
        items: List[Union[Model, BaseModel]],
        _relationships: FetchType,
    ) -> List[Model]:
        items = [i.model_copy() for i in items]
        await cls._load_relationships_for_items(items, _relationships)
        return items



from typing import Any, Dict, List, Optional, Sequence, Type, Union

from asyncpg import Connection, Pool, Record, connect, create_pool
from asyncpg.pool import PoolAcquireContext
from pypika.queries import QueryBuilder

from p3orm.drivers.base import BaseDriver
from p3orm.exceptions import AlreadyConnected, NotConnected
from p3orm.types import Model
from p3orm.utils import record_to_kwargs


class ConnectionContext:
    connection: Connection

    def __init__(self, connection: Connection):
        self.connection = connection

    async def __aenter__(self) -> Connection:
        return self.connection

    async def __aexit__(self, *exc):
        self.connection = None


class PoolConnectionContext(PoolAcquireContext):
    async def __aenter__(self) -> Connection:
        return await super().__aenter__()


class PostgresDriver(BaseDriver):
    connection: Connection
    pool: Pool

    def __init__(self):
        self.connection = None
        self.pool = None

    async def connect(
        self,
        dsn: str = None,
        *,
        user: str = None,
        password: str = None,
        database: str = None,
        host: str = None,
        port: int = None,
        **asyncpg_kwargs: Dict[str, Any],
    ):
        if self.is_connected():
            raise AlreadyConnected("A connection or pool is already established")

        self.pool = None
        self.connection = await connect(
            dsn=dsn,
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            **asyncpg_kwargs,
        )

    async def connect_pool(
        self,
        dsn: str = None,
        *,
        user: str = None,
        password: str = None,
        database: str = None,
        host: str = None,
        port: int = None,
        **asyncpg_kwargs: Dict[str, Any],
    ):
        if self.is_connected():
            raise AlreadyConnected("A databse connection or pool is already established")

        self.connection = None
        self.pool = await create_pool(
            dsn=dsn,
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            **asyncpg_kwargs,
        )

    async def disconnect(self):
        if self.connection:
            await self.connection.close()

        if self.pool:
            await self.pool.close()

        self.connection = None
        self.pool = None

    async def fetch_one(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> Optional[Model]:
        results = await self.fetch_many(cls, query, query_args=query_args)
        return None if len(results) == 0 else results[0]

    async def fetch_many(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> List[Model]:
        if isinstance(query, QueryBuilder):
            query = query.get_sql()

        if not self.is_connected():
            raise NotConnected("No database connection or pool is established")

        fetch_arguments = [query]

        if query_args:
            fetch_arguments += query_args

        results: List[Record] = []

        async with self._acquire_connection() as connection:
            results = await connection.fetch(*fetch_arguments)

        return [cls(**record_to_kwargs(row)) for row in results]

    def _acquire_connection(self) -> Union[ConnectionContext, PoolConnectionContext]:
        if self.connection:
            return ConnectionContext(self.connection)

        elif self.pool:
            return self.pool.acquire()

        raise NotConnected("No database connection or pool is established")

    def is_connected(self) -> bool:
        if self.connection and not self.connection.is_closed():
            return True

        if self.pool and not self.pool._closed:
            return True

        return False



from typing import Any, List, Optional, Sequence, Type, Union

from pypika.queries import QueryBuilder

from p3orm.types import Model


class BaseDriver:
    async def fetch_one(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> Optional[Model]:
        raise NotImplementedError

    async def fetch_many(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> List[Model]:
        raise NotImplementedError

    async def disconnect(self) -> None:
        raise NotImplementedError



from __future__ import annotations

from enum import Enum
from typing import NoReturn, Optional, Type, Union

from pypika import Field

from p3orm.exceptions import UnloadedRelationship
from p3orm.types import T


class UNLOADED:
    def __getattribute__(self, name) -> NoReturn:
        if name == "__class__":
            return UNLOADED

        raise UnloadedRelationship("_Relationship has not yet been loaded from the database")

    def __getitem__(*args) -> NoReturn:
        raise UnloadedRelationship("_Relationship has not yet been loaded from the database")

    def __eq__(self, other):
        return isinstance(other, UNLOADED)

    def __repr__(self) -> str:
        return f"<UNLOADED RELATIONSHIP>"


class _PormField(Field):
    pk: bool
    autogen: bool
    _type: Type
    field_name: str

    def __init__(
        self,
        _type: Type,
        column_name: Optional[str] = None,
        pk: Optional[bool] = False,
        autogen: Optional[bool] = False,
    ):
        self._type = _type
        self.column_name = column_name
        self.pk = pk
        self.autogen = autogen

        # This is used by underlying pypika logic, but column_name is used in p3orm for sanity
        self.name = column_name


def Column(
    _type: Type[T],
    column_name: Optional[str] = None,
    *,
    pk: Optional[bool] = False,
    autogen: Optional[bool] = False,
) -> Union[T, _PormField]:
    return _PormField(
        _type=_type,
        column_name=column_name,
        pk=pk,
        autogen=autogen,
    )


class RelationshipType(str, Enum):
    foreign_key = "foreign_key"
    reverse = "reverse"


class _Relationship:
    self_column: str
    foreign_column: str
    relationship_type: RelationshipType

    def __init__(self, self_column: str, foreign_column: str, relationship_type: RelationshipType):
        self.self_column = self_column
        self.foreign_column = foreign_column
        self.relationship_type = relationship_type


def ForeignKeyRelationship(self_column: str, foreign_column: str) -> _Relationship:
    return _Relationship(self_column, foreign_column, RelationshipType.foreign_key)


def ReverseRelationship(self_column: str, foreign_column: str) -> _Relationship:
    return _Relationship(self_column, foreign_column, RelationshipType.reverse)



from __future__ import annotations

import sqlite3
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union, get_args, get_origin

if TYPE_CHECKING:
    import asyncpg

from pypika.queries import QueryBuilder
from pypika.terms import (
    BasicCriterion,
    Comparator,
    ContainsCriterion,
    Criterion,
    Dialects,
    Equality,
    NullValue,
    Parameter,
    RangeCriterion,
)

from p3orm.exceptions import InvalidSQLiteVersion


class PormComparator(Comparator):
    empty = " "


def record_to_kwargs(record: asyncpg.Record) -> Dict[str, Any]:
    return {k: v for k, v in record.items()}


def with_returning(query: QueryBuilder, returning: Optional[str] = "*") -> str:
    return f"{query.get_sql()} RETURNING {returning}"


def paramaterize(
    criterion: Criterion, query_args: List[Any] = None, dialect: Dialects = None
) -> Tuple[Criterion, List[Any]]:
    if query_args == None:
        query_args = []

    param_start_index = max(len(query_args), 1)

    if isinstance(criterion, BasicCriterion):
        param = Parameter(f"${param_start_index}")
        query_args.append(criterion.right.value)
        return BasicCriterion(criterion.comparator, criterion.left, param, criterion.alias), query_args

    elif isinstance(criterion, ContainsCriterion):
        if dialect == Dialects.POSTGRESQL:
            param = Parameter(f"ANY (${param_start_index})")
            query_args.append([i.value for i in criterion.container.values if not isinstance(i, NullValue)])
            return BasicCriterion(Equality.eq, criterion.term, param, criterion.alias), query_args
        else:
            qs = ", ".join("?" for i in range(len(criterion.container.values)))
            param = Parameter(f"IN ({qs})")
            for i in criterion.container.values:
                if not isinstance(i, NullValue):
                    query_args.append(i.value)
                else:
                    query_args.append(None)
            return BasicCriterion(PormComparator.empty, criterion.term, param, criterion.alias), query_args

    elif isinstance(criterion, RangeCriterion):
        start_param = Parameter(f"${param_start_index}")
        end_param = Parameter(f"${param_start_index + 1}")

        query_args += [criterion.start.value, criterion.end.value]
        # There are several RangeCriterion, create a new one with the same subclass
        return criterion.__class__(criterion.term, start_param, end_param, criterion.alias), query_args

    return criterion, query_args


def is_optional(_type: Type):
    return get_origin(_type) is Union and type(None) in get_args(_type)


def validate_sqlite_version():
    if sqlite3.sqlite_version_info < (3, 35, 0):
        raise InvalidSQLiteVersion("p3orm requires SQLite engine version 3.35 or higher")



"""Utilitarian Python ORM for Postgres/SQLite, backed by asyncpg/aiosqlite, Pydantic, and PyPika"""

__version__ = "0.8.1"

# Provided for convenience
from pypika import Order

from .core import postgres, querybuilder, sqlite
from .exceptions import *
from .fields import *
from .table import *
from .types import *
from .utils import *



from __future__ import annotations

from typing import TYPE_CHECKING

from pypika.enums import Dialects
from pypika.queries import QueryBuilder

from p3orm.drivers.base import BaseDriver
from p3orm.exceptions import NotConnected
from p3orm.utils import validate_sqlite_version

if TYPE_CHECKING:
    from p3orm.drivers.postgres import PostgresDriver
    from p3orm.drivers.sqlite import SqliteDriver

DRIVER: BaseDriver = None
DIALECT: Dialects = None


def postgres() -> PostgresDriver:
    from p3orm.drivers.postgres import PostgresDriver

    global DRIVER, DIALECT
    if DRIVER:
        return DRIVER
    DRIVER = PostgresDriver()
    DIALECT = Dialects.POSTGRESQL
    return DRIVER


def sqlite() -> SqliteDriver:
    from p3orm.drivers.sqlite import SqliteDriver

    global DRIVER, DIALECT
    if DRIVER:
        return DRIVER
    validate_sqlite_version()
    DRIVER = SqliteDriver()
    DIALECT = Dialects.SQLLITE
    return DRIVER


def querybuilder() -> QueryBuilder:
    return QueryBuilder(dialect=DIALECT)


def driver() -> BaseDriver:
    global DRIVER
    if DRIVER:
        return DRIVER
    else:
        raise NotConnected("No driver as not connected")


def dialect() -> Dialects:
    global DIALECT
    if DIALECT:
        return DIALECT
    else:
        raise NotConnected("No dialect as not connected")



class PormException(Exception):
    ...


class MultipleResultsReturned(PormException):
    ...


class NoResultsReturned(PormException):
    ...


class MissingPrimaryKey(PormException):
    ...


class MultiplePrimaryKeys(PormException):
    ...


class MissingTablename(PormException):
    ...


class MissingRelationship(PormException):
    ...


class DatabaseException(PormException):
    ...


class InvalidRelationship(PormException):
    ...


class UnloadedRelationship(PormException):
    ...


class AlreadyConnected(PormException):
    ...


class NotConnected(PormException):
    ...


class InvalidSQLiteVersion(PormException):
    ...



import re
from pathlib import Path
from sqlite3 import Row
from typing import Any, Dict, List, Optional, Sequence, Type, Union

import aiosqlite
from aiosqlite.core import Connection
from pypika.queries import QueryBuilder

from p3orm.drivers.base import BaseDriver
from p3orm.exceptions import AlreadyConnected, NotConnected
from p3orm.types import Model


class SqliteDriver(BaseDriver):
    connection: Connection

    def __init__(self):
        self.connection = None

    async def connect(
        self,
        database: Union[str, Path],
        **aiosqlite_kwargs: Dict[str, Any],
    ):
        if self.is_connected():
            raise AlreadyConnected("A connection or pool is already established")

        self.connection = await aiosqlite.connect(database=database, **aiosqlite_kwargs)

    async def disconnect(self):
        await self.connection.close()
        self.connection = None

    async def fetch_one(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> Optional[Model]:
        results = await self.fetch_many(cls, query, query_args=query_args)
        return None if len(results) == 0 else results[0]

    async def fetch_many(
        self,
        cls: Type[Model],
        query: Union[str, QueryBuilder],
        query_args: Sequence[Any] = None,
    ) -> List[Model]:
        if isinstance(query, QueryBuilder):
            query = query.get_sql()

        if not self.is_connected():
            raise NotConnected("No database connection or pool is established")

        rows: List[Row] = []
        column_names: List[str] = []
        fields_map = {f.column_name: f.field_name for f in cls._fields()}

        # In a case such as an update, "where" arguments are placed toward the end
        # SQLite doesn't deal well with out of order params that don't match their
        # arguments, so here we shift the arguments a necessary amount so that they
        # match positionally rather than numerically
        params = re.findall(r"\$[0-9]+", query)
        if len(params) > 1:
            rotate_amount = len(params) - params.index("$1")
            for i in range(rotate_amount):
                query_args.append(query_args.pop(0))

        async with self.connection.execute(query, query_args) as cursor:
            rows = await cursor.fetchall()
            column_names = [column[0] for column in cursor.description]

            await self.connection.commit()

        return [
            cls(**{fields_map[column_name]: column_value for column_name, column_value in zip(column_names, row)})
            for row in rows
        ]

    def is_connected(self) -> bool:
        if not self.connection:
            return False

        if not (self.connection._running and self.connection._connection):
            return False

        return True
